import pytest
import sys
from rospy_yaml_include.yaml_include import RospyYamlInclude
import yaml
import os
import rospy

class TestRosParamInclude:
    """
    Class containing unit tests for dynamic include
    """

    @classmethod
    def setup_class(self):
        """
        setup function for tests
        """

        try:
            rospy.init_node("ros_param_include_test", log_level=rospy.INFO)
        except rospy.exceptions.ROSException:
            pass

        rospy.set_param("ros_param_include_test_param", "import")
        rospy.set_param("ros_param_include_test_param_circular", "circular_import")

    def test_relative_ros_param_include(self):
        """
        Unit test for ros param relative include
        """
        cwd = os.getcwd()

        yml = """
            value: 
            - 10
            fields: !ros_param_include test_files/${ros_param_include_test_param}.yaml
            """

        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

    def test_relative_ros_param_circular_include(self):
        """
        Unit test for circular includes in ros param relative include
        """
        cwd = os.getcwd()
        yml = """
            value: 
            - 10
            fields: !ros_param_include test_files/${ros_param_include_test_param_circular}.yaml
            """
        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            assert f"RecursionError caught: {error}"

    def test_absolute_ros_param_path_include(self):
        """
        Unit test for ros param absolute include
        """
        cwd = os.getcwd()

        yml = f"""
            value: 
            - 10
            fields: !ros_param_include {cwd}/test_files/${{ros_param_include_test_param}}.yaml
            """

        try:
            constructor = RospyYamlInclude()
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

    def test_absolute_ros_param_path_circular_include(self):
        """
        Unit test for circular includes in ros param absolute include
        """
        cwd = os.getcwd()
        yml = f"""
            value: 
            - 10
            fields: !ros_param_include {cwd}/test_files/${{ros_param_include_test_param_circular}}.yaml
            """
        try:
            constructor = RospyYamlInclude()
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            assert f"RecursionError caught: {error}"

if __name__ == "__main__":
    sys.exit(pytest.main(["--capture=no", __file__]))



import pytest
import sys
from rospy_yaml_include.yaml_include import RospyYamlInclude
import yaml
import os


class TestDynamicInclude:
    """
    Class containing unit tests for dynamic include
    """

    def test_relative_dynamic_include(self):
        """
        Unit test for dynamic relative include
        """
        cwd = os.getcwd()

        yml = """
            value: 
            - 10
            fields: !include test_files/import.yaml
            """

        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

    def test_relative_dynamic_circular_include(self):
        """
        Unit test for circular includes in dynamic relative include
        """
        cwd = os.getcwd()
        yml = """
            value: 
            - 10
            fields: !include test_files/circular_import.yaml
            """
        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            assert f"RecursionError caught: {error}"

    def test_dynamic_path_include(self):
        """
        Unit test for dynamic absolute include
        """
        cwd = os.getcwd()

        yml = f"""
            value: 
            - 10
            fields: !include {cwd}/test_files/import.yaml
            """

        try:
            constructor = RospyYamlInclude()
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

    def test_dynamic_path_circular_include(self):
        """
        Unit test for circular includes in dynamic absolute include
        """
        cwd = os.getcwd()
        yml = f"""
            value: 
            - 10
            fields: !include {cwd}/test_files/circular_import.yaml
            """
        try:
            constructor = RospyYamlInclude()
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            assert f"RecursionError caught: {error}"


if __name__ == "__main__":
    sys.exit(pytest.main(["--capture=no", __file__]))



import pytest
import sys
from rospy_yaml_include.yaml_include import RospyYamlInclude
import yaml
import os
import rospy

class TestRosParamSubstitute:
    """
    Class containing unit tests for dynamic include
    """

    def test_ros_param_substitute(self):
        """
        Unit test for dynamic relative include
        """

        try:
            rospy.init_node("ros_param_substitute_test", log_level=rospy.INFO)
        except rospy.exceptions.ROSException:
            pass

        rospy.set_param("ros_param_substitute_test_param", 10)

        cwd = os.getcwd()

        yml = """
            value: 
            - 10
            fields: !ros_param_substitute ${ros_param_substitute_test_param}
            """

        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

if __name__ == "__main__":
    sys.exit(pytest.main(["--capture=no", __file__]))



import pytest
import sys
from rospy_yaml_include.yaml_include import RospyYamlInclude
import yaml
import os

class TestVariableSubstitute:
    """
    Class containing unit tests for dynamic include
    """

    def test_variable_substitute(self):
        """
        Unit test for dynamic relative include
        """
        os.environ['TEST_SUBSTITUTE_VARIABLE'] = 'test'

        cwd = os.getcwd()

        yml = """
            value: 
            - 10
            fields: !variable_substitute ${TEST_SUBSTITUTE_VARIABLE}
            """

        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except Exception as error:
            pytest.fail(f"{error}")
        finally:
            del os.environ['TEST_SUBSTITUTE_VARIABLE']
if __name__ == "__main__":
    sys.exit(pytest.main(["--capture=no", __file__]))



import pytest
import sys
from rospy_yaml_include.yaml_include import RospyYamlInclude
import yaml
import os


class TestVariableInclude:
    """
    Class containing unit tests for dynamic include
    """

    @classmethod
    def setup_class(self):
        """
        setup function for tests
        """
        os.environ['TEST_INCLUDE_VARIABLE'] = 'import'
        os.environ['TEST_INCLUDE_VARIABLE_CIRCULAR'] = 'circular_import'

    def test_relative_variable_include(self):
        """
        Unit test for variable relative include
        """
        cwd = os.getcwd()

        yml = """
            value: 
            - 10
            fields: !variable_include test_files/${TEST_INCLUDE_VARIABLE}.yaml
            """

        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

    def test_relative_variable_circular_include(self):
        """
        Unit test for circular includes in variable relative include
        """
        cwd = os.getcwd()
        yml = """
            value: 
            - 10
            fields: !variable_include test_files/${TEST_INCLUDE_VARIABLE_CIRCULAR}.yaml
            """
        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            assert f"RecursionError caught: {error}"

    def test_absolute_variable_path_include(self):
        """
        Unit test for variable absolute include
        """
        cwd = os.getcwd()

        yml = f"""
            value: 
            - 10
            fields: !variable_include {cwd}/test_files/${{TEST_INCLUDE_VARIABLE}}.yaml
            """

        try:
            constructor = RospyYamlInclude()
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")

    def test_absolute_variable_path_circular_include(self):
        """
        Unit test for circular includes in variable absolute include
        """
        cwd = os.getcwd()
        yml = f"""
            value: 
            - 10
            fields: !variable_include {cwd}/test_files/${{TEST_INCLUDE_VARIABLE_CIRCULAR}}.yaml
            """
        try:
            constructor = RospyYamlInclude()
            yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            assert f"RecursionError caught: {error}"

    @classmethod
    def teardown_class(self):
        """
        teardown function for tests
        """

        del os.environ['TEST_INCLUDE_VARIABLE']

if __name__ == "__main__":
    sys.exit(pytest.main(["--capture=no", __file__]))



"""
rospy_yaml_include
------------------

This module provides a way to include yaml files in other yaml files.
"""
from ast import literal_eval
import os
import re
from typing import Any, Union, Dict, List, Tuple

import rospkg
import rospy
import yaml


def convert_to_type(default_value: str) -> Any:
    """
    Converts a <str> input to its literal <type> output
    """
    try:  # to convert str to Dict, List, int, float, bool
        return literal_eval(default_value)
    except (ValueError, SyntaxError):
        return default_value


def extract_arguments(input_string: str,
                      pattern: str = r'\$\{([^}]+)\}') \
        -> List[Tuple[str, str]]:
    matches = re.findall(pattern, input_string)
    params = []
    for match in matches:
        match = match.split(';')
        param = match[0].strip()
        try:
            default = match[1].strip()
        except IndexError:
            default = None
        params.append((param, default))
    return params


class RospyYamlInclude:
    """
    RospyYamlInclude class
    """

    def __init__(
        self, loader: type = yaml.SafeLoader, base_directory: str = None, import_limit: int = 150
    ) -> None:
        self.import_limit = import_limit
        self.import_count = 0

        self.loader = loader
        self.base_directory = base_directory

    class _RosInclude:
        """
        Mappping for !ros_include constructor
        """

        def __init__(self, package, extension) -> None:
            self.package = package
            self.extension = extension

    def _ros_include_constructor(
        self, loader: type, node: yaml.nodes.MappingNode
    ) -> dict:
        """
        _ros_include_constructor function handles !ros_include tag
        """
        self.import_count += 1
        if self.import_count > self.import_limit:
            raise RecursionError(
                "Maximum import limit reached, check for circular references or increase import limit"
            )

        rospack = rospkg.RosPack()

        file = self._RosInclude(**loader.construct_mapping(node))

        include_file = os.path.join(
            rospack.get_path(file.package),
            file.extension,
        )

        with open(include_file, encoding="utf-8") as yaml_file:
            return yaml.load(yaml_file, Loader=self.add_constructor())

    def _path_include_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> dict:
        """
        _path_include_constructor function handles !path_include tag

        """
        self.import_count += 1
        if self.import_count > self.import_limit:
            raise RecursionError(
                "Maximum import limit reached, check for circular references or increase import limit"
            )

        file = loader.construct_scalar(node)

        with open(file, encoding="utf-8") as yaml_file:
            return yaml.load(yaml_file, Loader=self.add_constructor())

    def _relative_include_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> dict:
        """
        _path_include_constructor function handles !relative_include tag

        this can be used to import a yaml relative to a base directory provided in the class init
        """

        self.import_count += 1
        if self.import_count > self.import_limit:
            raise RecursionError(
                "Maximum import limit reached, check for circular references or increase import limit"
            )

        if self.base_directory is None:
            raise ValueError(
                "base_directory must be provided in class init to use !relative_include"
            )

        file = loader.construct_scalar(node)

        include_file = os.path.join(
            self.base_directory,
            file,
        )

        with open(include_file, encoding="utf-8") as yaml_file:
            return yaml.load(yaml_file, Loader=self.add_constructor())

    def _dynamic_include_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> dict:
        """
        _dynamic_include_constructor function handles !include tag

        this constructor attempts to infer the type of include based on the file extension
        """

        self.import_count += 1
        if self.import_count > self.import_limit:
            raise RecursionError(
                "Maximum import limit reached, check for circular references or increase import limit"
            )

        file = loader.construct_scalar(node)

        if file.startswith("/"):
            include_file = file
        else:
            if self.base_directory is None:
                raise ValueError(
                    "base_directory must be provided in class init to use relative include"
                )

            include_file = os.path.join(
                self.base_directory,
                file,
            )

        with open(include_file, encoding="utf-8") as yaml_file:
            return yaml.load(yaml_file, Loader=self.add_constructor())

    def _variable_subsitute_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> dict:
        """
        _variable_subsitute_constructor function handles !variable_substitute tag

        this can be used to substitute a variable in a yaml file
        """

        param_string = loader.construct_scalar(node)
        variables = re.findall(r"\${(.*?)}", param_string)
        for variable in variables:
            fill_param = os.getenv(variable, None)
            if fill_param is not None:
                param_string = param_string.replace(f"${{{variable}}}", str(fill_param))
            else:
                raise ValueError(f"env {variable} not found")

        return param_string

    def _variable_include_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> dict:
        """
        _variable_include_constructor function handles !variable_include tag

        this can be used to import a yaml while substituting a env variable
        """

        self.import_count += 1
        if self.import_count > self.import_limit:
            raise RecursionError(
                "Maximum import limit reached, check for circular references or increase import limit"
            )

        file = loader.construct_scalar(node)
        variables = re.findall(r"\${(.*?)}", file)
        for variable in variables:
            fill_param = os.getenv(variable, None)
            if fill_param is not None:
                file = file.replace(f"${{{variable}}}", str(fill_param))
            else:
                raise ValueError(f"env {variable} not found")

        if file.startswith("/"):
            include_file = file
        else:
            if self.base_directory is None:
                raise ValueError(
                    "base_directory must be provided in class init to use relative include"
                )

            include_file = os.path.join(
                self.base_directory,
                file,
            )

        with open(include_file, encoding="utf-8") as yaml_file:
            return yaml.load(yaml_file, Loader=self.add_constructor())
        
    def _ros_param_substitute_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> Union[None, Dict, List, bool, int, float, str]:
        """
        _ros_param_substitute_constructor function handles !ros_param_substitute tag

        this can be used to import a yaml substitute a rosparam

        includes option for "; <default value>" to provide a default value if rosparam is not found
            MUST be separated by semicolon ';' to allow for dict, list as default values
        """
        param_string = loader.construct_scalar(node)
        rospy.logdebug(f'ROS_PARAM_SUBSTITUTE STR: {param_string}')

        # pattern = r'\$\{([^}]+)\}'  # finds ${<ros_param_substitute>; <default>}
        # the pattern below is a modified version of the above pattern,
        # allowing for nested {} in the default value e.g.:
        # https://regex101.com/r/wPCbj8/1
        pattern = r'\$\{([^{}]*(?:{[^{}]*}[^{}]*)*)\}'  # finds ${<ros_param_substitute>; <{default}>}

        params = extract_arguments(param_string, pattern=pattern)
        rospy.logdebug(f'PARAMS [(param, default),]: {params}')

        clean_param_string = [f'${{{param}}}' for param, default in params]

        # Join the matches to re-form the resultant param_string (with defaults stripped)
        param_string = ' '.join(clean_param_string)
        rospy.logdebug(f'ROS_PARAM_SUBSTITUTE STR (CLEAN): {param_string}')

        for param, default in params:
            # if param is not found, use default value (if provided, else None)
            fill_param = rospy.get_param(param, default)

            # immediately throw error if param is not found AND default is not provided
            if fill_param is None:
                rospy.logerr(f"ERROR: tag= {node.tag} value='{node.value}'")
                raise ValueError(f"rosparam '{param}' not found; no default provided "
                                 f"[tag= {node.tag} value='{node.value}']")

            # TODO: determine if we want this capability?
            # check if default value is a rosparam substitution
            if isinstance(fill_param, str) and fill_param.startswith('${'):
                default_param_sub = re.findall(r"\${(.*?)}", fill_param)[0]
                fill_param = rospy.get_param(default_param_sub, None)

            rospy.logdebug(f'FILL_PARAM: {type(fill_param)} {fill_param}')

            param_string = param_string.replace(f"${{{param}}}", str(fill_param))
            rospy.logdebug(f"PARAM_STRING: '{param_string}'")

        # param_string Union[None, Dict, List, bool, int, float, str]
        return convert_to_type(param_string)

    def _ros_param_include_constructor(
        self, loader: type, node: yaml.nodes.ScalarNode
    ) -> dict:
        """
        _ros_param_include_constructor function handles !ros_param_include tag

        this can be used to import a yaml with a rosparam substitution
        """

        self.import_count += 1
        if self.import_count > self.import_limit:
            raise RecursionError(
                "Maximum import limit reached, check for circular references or increase import limit"
            )

        file = loader.construct_scalar(node)
        params = re.findall(r"\${(.*?)}", file)
        for param in params:
            fill_param = rospy.get_param(param, None)
            if fill_param is not None:
                file = file.replace(f"${{{param}}}", str(fill_param))
            else:
                raise ValueError(f"rosparam {param} not found")

        if file.startswith("/"):
            include_file = file
        else:
            if self.base_directory is None:
                raise ValueError(
                    "base_directory must be provided in class init to use relative include"
                )

            include_file = os.path.join(
                self.base_directory,
                file,
            )

        with open(include_file, encoding="utf-8") as yaml_file:
            return yaml.load(yaml_file, Loader=self.add_constructor())
        
    def add_constructor(self) -> type:
        """
        add constructor to yaml
        """

        loader = self.loader
        loader.add_constructor("!ros_include", self._ros_include_constructor)
        loader.add_constructor("!path_include", self._path_include_constructor)
        loader.add_constructor("!relative_include", self._relative_include_constructor)
        loader.add_constructor("!include", self._dynamic_include_constructor)
        loader.add_constructor("!variable_substitute", self._variable_subsitute_constructor)
        loader.add_constructor("!variable_include", self._variable_include_constructor)
        loader.add_constructor("!ros_param_substitute", self._ros_param_substitute_constructor)
        loader.add_constructor("!ros_param_include", self._ros_param_include_constructor)

        return loader



#!/usr/bin/env python3
import json
import os
import sys

import pytest
import rospy
import yaml

from rospy_yaml_include.yaml_include import (RospyYamlInclude,
                                             extract_arguments,
                                             )


def run(inputs):
    for idx, ip in enumerate(inputs):
        print(f'For Input {idx+1}: "{ip}"')
        results = extract_arguments(ip)
        for r in [x for x in results]:
            print(f'  -> {r}')


class TestRosParamSubsExtractingDefaults:
    """
    Class containing unit tests for ros_param_substitute include
    with default value inputs
    """
    def test_extracting_default_args(self):
        """
        Unit test for dynamic relative include
        """
        try:
            rospy.init_node("ros_param_substitute_test_defaults", log_level=rospy.DEBUG)
        except rospy.exceptions.ROSException:
            pass

        # NOTE: commenting lines MIGHT require restarting roscore
        rospy.set_param("/test", 1)  # comment to raise ValueError
        rospy.set_param("test2", 2)  # comment to test w/ default value "{name}/status"
        rospy.set_param("test3", [3, 3])
        rospy.set_param("test7", [7, 8, 9])  # comment to test w/ default value "[7,7,8,8,9,9]"

        yml = """
            value: !ros_param_substitute ${/test}
            value_default: !ros_param_substitute ${test2; "{name}/status"}
            fields: !ros_param_substitute ${test3;5}
            fields_augmented: !ros_param_substitute ${test3; 5} ${test4; 50}
            fields_default_dict: !ros_param_substitute ${test4; {"test5":5, "test6":6.0,"hello":"world"}}
            fields_default_list: !ros_param_substitute ${test7; [7, 7, 8, 8, 9, 9]}
            """

        cwd = os.getcwd()

        # TODO:
        #  - [ ] better handling of fields_augmented key-value substitutions
        #  - [ ] better handling of fields_default_dict key-value
        #           substitution (yaml.scanner.ScannerError)
        try:
            constructor = RospyYamlInclude(base_directory=cwd)
            params = yaml.load(yml, Loader=constructor.add_constructor())
        except RecursionError as error:
            pytest.fail(f"RecursionError: Maximum recursion limit reached: {error}")
        except yaml.scanner.ScannerError as error:
            pytest.fail(f"ScannerError: {error} (try removing spaces between "
                        f"any mappings [e.g. 'k':v] in the default input)")
        else:
            rospy.loginfo(f'Pretty Print:\n{json.dumps(params, indent=4)}')
            rospy.loginfo(params)


if __name__ == "__main__":
    # Example extract_arguments() output:
    inputs = [
        "${~status_max_queuesize} ${~status_priority}",
        "${~status_max_queuesize;30} ${~status_priority;50}",
        "${~status_max_queuesize; 30} ${~status_priority}",
        "${~status_max_queuesize} ${~status_priority; 5",
    ]
    run(inputs)

    # Example TestExtractingDefaultArgs
    test = TestRosParamSubsExtractingDefaults()
    test.test_extracting_default_args()
    sys.exit(pytest.main(["--capture=no", __file__]))
